"""
Shared ETL (extract, transform, load) code for fetching and loading data from HCA DSS.

The main entry point is the DSSExtractor.extract() function, which takes optional transform and load callbacks that
will be invoked for each bundle fetched.

Example usage:

    from dcplib.etl import DSSExtractor

    def tf(*args, **kwargs):
        print("Transformer", args, kwargs)
        return "TEST"

    def ld(*args, **kwargs):
        print("Loader", args, kwargs)

    def fn(*args, **kwargs):
        print("Finalizer", args, kwargs)

    DSSExtractor(staging_directory=".").extract(transformer=tf, loader=ld, finalizer=fn)
"""

import os, sys, json, concurrent.futures, hashlib, logging, threading
from collections import defaultdict
from fnmatch import fnmatchcase

import hca
from ..networking import http
from hca.util import RetryPolicy

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

class DSSExtractor:
    default_bundle_query = {'query': {'bool': {'must_not': {'term': {'admin_deleted': True}}}}}
    default_content_type_patterns = ['application/json; dcp-type="metadata*"']

    def __init__(self, staging_directory, content_type_patterns: list = None, filename_patterns: list = None,
                 dss_client: hca.dss.DSSClient = None):
        self.sd = staging_directory
        self.content_type_patterns = content_type_patterns or self.default_content_type_patterns
        self.filename_patterns = filename_patterns or []
        self.b2f = defaultdict(set)
        self.staged_bundles = []
        self._dss_client = dss_client or hca.dss.DSSClient()
        self._dss_swagger_url = None

    # concurrent.futures.ProcessPoolExecutor requires objects to be picklable.
    # hca.dss.DSSClient is unpicklable and is stubbed out here to preserve DSSExtractor's picklability.
    def __getstate__(self):
        state = dict(self.__dict__)
        state["_dss_swagger_url"] = self.dss_client.swagger_url
        state["_dss_client"] = None
        return state

    @property
    def dss_client(self):
        if self._dss_client is None:
            self._dss_client = hca.dss.DSSClient(swagger_url=self._dss_swagger_url)
        return self._dss_client

    def link_file(self, bundle_uuid, bundle_version, f):
        if not os.path.exists(f"{self.sd}/bundles/{bundle_uuid}.{bundle_version}/{f['name']}"):
            logger.debug("Linking fetched file %s/%s", bundle_uuid, f["name"])
            os.symlink(f"../../files/{f['uuid']}", f"{self.sd}/bundles/{bundle_uuid}.{bundle_version}/{f['name']}")

    def get_file(self, f, bundle_uuid, bundle_version, print_progress=True):
        logger.debug("[%s] Fetching %s:%s", threading.current_thread().getName(), bundle_uuid, f["name"])
        res = http.get(f"{self.dss_client.host}/files/{f['uuid']}", params={"replica": "aws", "version": f["version"]})
        res.raise_for_status()
        with open(f"{self.sd}/files/{f['uuid']}", "wb") as fh:
            fh.write(res.content)
        self.link_file(bundle_uuid, bundle_version, f)
        logger.debug("Wrote %s:%s", bundle_uuid, f["name"])
        if print_progress:
            sys.stdout.write(".")
            sys.stdout.flush()
        return f, bundle_uuid, bundle_version

    def should_fetch_file(self, f):
        if any(fnmatchcase(f["content-type"], p) for p in self.content_type_patterns):
            return True
        if any(fnmatchcase(f["name"], p) for p in self.filename_patterns):
            return True
        return False

    def get_files_to_fetch_for_bundle(self, bundle_uuid, bundle_version):
        logger.debug("[%s] Fetching manifest for bundle %s", threading.current_thread().getName(), bundle_uuid)
        res = http.get(f"{self.dss_client.host}/bundles/{bundle_uuid}", params={"replica": "aws"})
        res.raise_for_status()
        logger.debug("Scanning bundle %s", bundle_uuid)
        files_to_fetch = []
        for f in res.json()["bundle"]["files"]:
            if self.should_fetch_file(f):
                os.makedirs(f"{self.sd}/bundles/{bundle_uuid}.{bundle_version}", exist_ok=True)
                try:
                    with open(f"{self.sd}/files/{f['uuid']}", "rb") as fh:
                        file_csum = hashlib.sha256(fh.read()).hexdigest()
                        if file_csum == f["sha256"]:
                            self.link_file(bundle_uuid, bundle_version, f)
                            continue
                except (FileNotFoundError, ):
                    pass
                files_to_fetch.append(f)
            else:
                logger.debug("Skipping file %s/%s (no filter match)", bundle_uuid, f["name"])
        return bundle_uuid, bundle_version, files_to_fetch

    def enqueue_bundle_manifests(self, percent_complete, bundles_complete, total_bundles, f2f_futures, ff_futures,
                                 executor):
        logger.info("[%d%%] [%d/%d] Processing bundle batch",
                    percent_complete, bundles_complete, total_bundles)
        for future in concurrent.futures.as_completed(f2f_futures):
            bundle_uuid, bundle_version, files_to_fetch = future.result()
            if files_to_fetch:
                logger.info("[%d%%] Fetching %d files for bundle %s @ %s",
                            percent_complete, len(files_to_fetch), bundle_uuid, bundle_version)
                for f in files_to_fetch:
                    ff_futures.append(executor.submit(self.get_file, f, bundle_uuid, bundle_version))
                    self.b2f[bundle_uuid + "." + bundle_version].add(f["name"])
            else:
                self.staged_bundles.append([bundle_uuid, bundle_version])
        return len(f2f_futures)

    def extract(self, query: dict = None, transformer: callable = None, loader: callable = None,
                finalizer: callable = None, max_workers=512, max_dispatchers=1,
                dispatch_executor_class: concurrent.futures.Executor = concurrent.futures.ThreadPoolExecutor):
        if query is None:
            query = self.default_bundle_query
        os.makedirs(f"{self.sd}/files", exist_ok=True)
        os.makedirs(f"{self.sd}/bundles", exist_ok=True)
        f2f_futures, ff_futures = [], []
        bundles_complete, percent_complete = 0, -1
        with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
            total_bundles = self.dss_client.post_search(es_query=query, replica="aws")["total_hits"]
            logger.info("Scanning %s bundles", total_bundles)
            for b in self.dss_client.post_search.iterate(es_query=query, replica="aws", per_page=500):
                bundle_uuid, bundle_version = b["bundle_fqid"].split(".", 1)
                f2f_futures.append(executor.submit(self.get_files_to_fetch_for_bundle, bundle_uuid, bundle_version))
                if len(f2f_futures) > max_workers * 2:
                    bundles_complete += self.enqueue_bundle_manifests(percent_complete, bundles_complete, total_bundles,
                                                                      f2f_futures, ff_futures, executor)
                    f2f_futures.clear()

                if int((bundles_complete / total_bundles) * 100) > percent_complete:
                    percent_complete = int((bundles_complete / total_bundles) * 100)
                    logger.info("%d%% complete (%d/%d bundles)", percent_complete, bundles_complete, total_bundles)
            bundles_complete += self.enqueue_bundle_manifests(percent_complete, bundles_complete, total_bundles,
                                                              f2f_futures, ff_futures, executor)

        with dispatch_executor_class(max_workers=max_dispatchers) as executor:
            for future in concurrent.futures.as_completed(ff_futures):
                f, bundle_uuid, bundle_version = future.result()
                self.b2f[bundle_uuid + "." + bundle_version].remove(f["name"])
                if len(self.b2f[bundle_uuid + "." + bundle_version]) == 0:
                    executor.submit(self.dispatch_callbacks, bundle_uuid, bundle_version, transformer, loader)
                else:
                    logger.debug("%s: %d files to go", bundle_uuid, len(self.b2f[bundle_uuid + "." + bundle_version]))
            for bundle_uuid, bundle_version in self.staged_bundles:
                executor.submit(self.dispatch_callbacks, bundle_uuid, bundle_version, transformer, loader)

        if finalizer is not None:
            finalizer(extractor=self)
        logger.info("Done (%d/%d bundles)", bundles_complete, total_bundles)

    def dispatch_callbacks(self, bundle_uuid, bundle_version, transformer, loader):
        bundle_path = f"{self.sd}/bundles/{bundle_uuid}.{bundle_version}"
        if transformer is not None:
            tb = transformer(bundle_uuid=bundle_uuid, bundle_version=bundle_version, bundle_path=bundle_path,
                             extractor=self)
        if loader is not None:
            loader(extractor=self, transformer=transformer, bundle=tb)



import os
import unittest

from dcplib.media_types import DcpMediaType


class TestDcpMediaType(unittest.TestCase):

    BUNDLE_FIXTURE_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), 'fixtures'))

    TEST_FILES_AND_STRINGS = {
        'assay.json': 'application/json; dcp-type="metadata/assay"',
        'project.json': 'application/json; dcp-type="metadata/project"',
        'sample.json': 'application/json; dcp-type="metadata/sample"',
        'manifest.json': 'application/json; dcp-type=metadata',
        'SRR2963332_1.fastq.gz': 'application/gzip; dcp-type=data',
        'demo.gene.counts.txt': 'text/plain; dcp-type=data',
        'mystery_file': 'application/octet-stream; dcp-type=data'
    }

    def test_file_recognition(self):
        for filename, media_type in TestDcpMediaType.TEST_FILES_AND_STRINGS.items():
            file_path = os.path.join(self.BUNDLE_FIXTURE_DIR, filename)
            self.assertEqual(str(DcpMediaType.from_file(file_path)), media_type)



import mimetypes
import operator
import re

import puremagic


class MediaType(object):

    RFC7320_TOKEN_CHARSET = r"!#$%&'*+\-.^_`|~\w"
    _TYPE_REGEX = re.compile(
        r"(?P<top_level_type>.+)"  # 'application'
        r"/"                       # '/'
        r"(?P<subtype>[^+]+)"      # 'json'
        r"("                       # stat of suffix group
        r"\+(?P<suffix>.+)"        # '+zip'
        r")?"                      # end optional suffix group
    )
    _PARAM_REGEX = re.compile(
        r"\s*"                               # optional whitespace
        r"(?P<key>"                          # start of group 'key'
        r"[" + RFC7320_TOKEN_CHARSET + "]+"  # key
        r")"                                 # end of group 'key'
        r"\s*=\s*"                           # '='
        r"(?P<value>"                        # start of group 'value'
        r'"(?:[^\\"]|\\.)*"'                 # Any doublequoted string
        r"|"                                 # or
        "[" + RFC7320_TOKEN_CHARSET + "]*"   # A word of token-chars (doesn't need quoting)
        r")"                                 # end of group 'value'
        r"\s*(;|$)"                          # Ending either at semicolon, or EOS.
    )
    _QUOTED_PAIR_REGEX = re.compile(r"[\\].")

    @classmethod
    def from_string(cls, media_type):
        first_semi_position = media_type.find(';')
        if first_semi_position == -1:
            type_info = media_type
            parameters = {}
        else:
            type_info = media_type[0:first_semi_position]
            parameters = MediaType._parse_parameters(media_type[first_semi_position + 1:])

        type_info = MediaType._TYPE_REGEX.match(type_info)

        return cls(
            type_info.group('top_level_type'), type_info.group('subtype'),
            suffix=type_info.group('suffix'), parameters=parameters)

    @classmethod
    def from_file(cls, file_path):
        media_type_string = cls._media_type_of_file(file_path)
        media_type = cls.from_string(media_type_string)
        return media_type

    def __init__(self, top_level_type, subtype, suffix=None, parameters={}):
        self.top_level_type = top_level_type
        self.subtype = subtype
        self.suffix = suffix
        self.parameters = parameters

    def __str__(self):
        result = [self.main_type]
        if self.parameters:
            sorted_params = sorted(self.parameters.items(), key=operator.itemgetter(0))
            for k, v in sorted_params:
                result.append('{key}={value}'.format(key=k, value=self._minimally_quoted_parameter_value(v)))
        return '; '.join(result)

    @property
    def main_type(self):
        maintype = "{type}/{subtype}".format(type=self.top_level_type, subtype=self.subtype)
        if self.suffix:
            maintype += "+" + self.suffix
        return maintype

    @staticmethod
    def _parse_parameters(params_string):
        parameters = {}
        cursor = 0
        string_length = len(params_string)

        while 0 <= cursor < string_length:
            match = MediaType._PARAM_REGEX.match(params_string, cursor)
            if not match:
                break
            value = match.group('value')
            parameters[match.group('key')] = MediaType._unquote(value)
            cursor = match.end(0)
        return parameters

    @staticmethod
    def _media_type_of_file(file_path):
        return MediaType._puremagic_media_type_of_file(file_path) or \
            MediaType._mimetypes_media_type_of_file(file_path) or \
            "application/octet-stream"

    @staticmethod
    def _puremagic_media_type_of_file(file_path):
        array_of_type_arrays = puremagic.magic_file(file_path)
        types = list(map(lambda x: x[1], array_of_type_arrays))
        types = list(filter(lambda x: x and x != '', types))
        if len(types) == 0:
            return None
        media_type = types[0]
        if media_type == 'application/x-ipynb+json':  # All JSON files are not Jupiter notebooks
            media_type = 'application/json'
        elif media_type == 'application/x-gzip':  # deprecated
            media_type = 'application/gzip'
        return media_type

    @staticmethod
    def _mimetypes_media_type_of_file(file_path):
        media_type, encoding = mimetypes.guess_type(file_path)
        return media_type

    @staticmethod
    def _minimally_quoted_parameter_value(value):
        """
        Per RFC 7321 (https://tools.ietf.org/html/rfc7231#section-3.1.1.1):
        Parameters values don't need to be quoted if they are a "token".
        Token characters are defined by RFC 7320 (https://tools.ietf.org/html/rfc7230#section-3.2.6).
        Otherwise, parameters values can be a "quoted-string".
        So we will quote values that contain characters other than the standard token characters.
        """
        if re.match("^[{charset}]*$".format(charset=MediaType.RFC7320_TOKEN_CHARSET), value):
            return value
        else:
            return MediaType._quote(value)

    _Translator = {
        '"': '\\"',
        '\\': '\\\\'
    }

    @staticmethod
    def _quote(string):
        """
        https://tools.ietf.org/html/rfc7230#section-3.2.6

        The backslash octet ("\") can be used as a single-octet quoting
        mechanism within quoted-string and comment constructs.  Recipients
        that process the value of a quoted-string MUST handle a quoted-pair
        as if it were replaced by the octet following the backslash.

          quoted-pair    = "\" ( HTAB / SP / VCHAR / obs-text )

        A sender SHOULD NOT generate a quoted-pair in a quoted-string except
        where necessary to quote DQUOTE and backslash octets occurring within
        that string.
        """
        return '"' + ''.join(map(MediaType._Translator.get, string, string)) + '"'

    @staticmethod
    def _unquote(string):
        # If there aren't any doublequotes, then there can't be any special characters.
        if len(string) < 2:
            return string
        if string[0] != '"' or string[-1] != '"':
            return string

        string = string[1:-1]  # Remove the "s

        # Check for special sequences.  Examples:
        #    \"   --> "
        cursor = 0
        strlen = len(string)
        result = []
        while 0 <= cursor < strlen:
            match = MediaType._QUOTED_PAIR_REGEX.search(string, cursor)
            if match:
                backslash_location = match.start(0)
                result.append(string[cursor:backslash_location])
                result.append(string[backslash_location + 1])
                cursor = backslash_location + 2
            else:
                result.append(string[cursor:])
                break

        return ''.join(result)



#!/usr/bin/env python
# coding: utf-8

from __future__ import absolute_import, division, print_function, unicode_literals

import os
import sys
import unittest

pkg_root = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))  # noqa
sys.path.insert(0, pkg_root)  # noqa

from dcplib import s3_multipart
from dcplib.checksumming_io import ChecksummingSink
from . import TEST_FILE, TEST_FILE_CHECKSUMS


class TestChecksummingSink(unittest.TestCase):
    file_size = os.path.getsize(TEST_FILE)
    chunk_size = s3_multipart.get_s3_multipart_chunk_size(file_size)

    def check_sums(self, checksums):
        [self.assertEqual(checksums[hash_function].lower(), TEST_FILE_CHECKSUMS[hash_function].lower()) for
         hash_function in checksums.keys()]

    def test_crc32c_calculation(self):
        crc32 = ChecksummingSink.CRC32C()
        checksum = None

        with open(TEST_FILE, 'rb') as fh:
            data = fh.read()
            crc32.update(data)
            checksum = crc32.hexdigest()

        self.assertEqual(checksum.lower(), TEST_FILE_CHECKSUMS['crc32c'].lower())

    def test_checksums_after_single_write(self):
        sink = ChecksummingSink(self.chunk_size)
        with open(TEST_FILE, 'rb') as fh:
            data = fh.read()
            sink.write(data)
        sums = sink.get_checksums()
        self.check_sums(sums)

    def test_checksums_after_multiple_write(self):
        sink = ChecksummingSink(self.chunk_size)
        with open(TEST_FILE, 'rb') as fh:
            while True:
                data = fh.read(self.chunk_size)
                if not data:
                    break
                sink.write(data)
        sums = sink.get_checksums()
        self.check_sums(sums)

    def test_hash_function_list_is_configurable(self):
        checksums_to_compute = ['sha1', 's3_etag']
        sink = ChecksummingSink(self.chunk_size, hash_functions=checksums_to_compute)
        with open(TEST_FILE, 'rb') as fh:
            data = fh.read()
            sink.write(data)
        sums = sink.get_checksums()
        self.assertEqual(list(sorted(sums.keys())), sorted(checksums_to_compute))
        [self.assertEqual(TEST_FILE_CHECKSUMS[checksum].lower(), sums[checksum].lower()) for checksum in
         checksums_to_compute]


if __name__ == '__main__':
    unittest.main()



import json
import os

from .aws_secret import AwsSecret


class Config:

    """
    This Config class stores its configuration in an AwsSecret.
    Subclass it to create your config, as follows (Python 3.6):

      class MyComponentConfig(Config):

          def __init__(self, *args, **kwargs):
              super().__init__('my_component', **kwargs)

    Setup your secrets in AWS Secrets Manager (e.g. use Terraform), as a JSON hash,
    at path:

        dcp/my_component/<deployment>/secrets

    To access secrets, (assuming that 'my_secret' is a key of your JSON hash):

        MyComponentConfig().my_secret

    Implements singleton-like behavior using techniques from
    http://python-3-patterns-idioms-test.readthedocs.io/en/latest/Singleton.html.
    All instances of this class will share the same config data.
    If you subclass this class, the subclass gets its own data, but all
    instances of the subclass share that data.
    """

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, '_config'):
            cls.reset()
        return super(Config, cls).__new__(cls)

    """
    If source is specified, it must be the path to a JSON file
    """
    def __init__(self, component_name, deployment=None, source=None, secret_name='secrets'):
        super(Config, self).__init__()
        self._component_name = component_name
        self._deployment = deployment or os.environ['DEPLOYMENT_STAGE']
        self._secret_name = secret_name
        self._source = self._determine_source(source)

    @property
    def config(self):
        return self.__class__._config

    def __getattr__(self, name):
        if self.config_is_loaded():
            return self.value_from_config(name) \
                or self.value_from_env(name) \
                or self.raise_error(name)
        else:
            return self.value_from_env(name) \
                or (self.load() and self.value_from_config(name)) \
                or self.raise_error(name)

    @classmethod
    def reset(cls):
        cls._config = None
        cls.use_env = False

    """
    Bypass the load mechanism and set secrets directly.
    Used in testing.
    """
    def set(self, config):
        self.__class__._config = config
        self.__class__.use_env = False

    def load(self):
        if self._source == 'aws':
            self.load_from_aws()
        else:
            self.load_from_file(self._source)
        return True  # so we can be used in 'and' statements

    def config_is_loaded(self):
        return self.config is not None

    def load_from_aws(self):
        secret_path = "dcp/{component_name}/{deployment}/{secret_name}".format(
            component_name=self._component_name, deployment=self._deployment, secret_name=self._secret_name)
        secret = AwsSecret(secret_path)
        self.from_json(secret.value)

    def load_from_file(self, config_file_path):
        with open(config_file_path, 'r') as config_fp:
            self.from_json(config_fp.read())

    def from_json(self, config_json):
        self.__class__._config = json.loads(config_json)

    def _determine_source(self, source):
        if source:
            pass
        elif 'CONFIG_SOURCE' in os.environ:
            source = os.environ['CONFIG_SOURCE']
        else:
            source = 'aws'
        return source

    def value_from_config(self, name):
        if name in self.config:
            return self.config[name]
        else:
            return None

    def value_from_env(self, name):
        if self.__class__.use_env and name.upper() in os.environ:
            return os.environ[name.upper()]
        else:
            return None

    def raise_error(self, name):
        raise RuntimeError(name + " is not in configuration")



import os

from .media_type import MediaType


class DcpMediaType(MediaType):

    DCP_METADATA_JSON_FILES = ('assay', 'project', 'sample')

    @classmethod
    def from_file(cls, file_path, dcp_type=None):
        media_type = super(DcpMediaType, cls).from_file(file_path)
        if dcp_type:
            media_type.parameters['dcp-type'] = dcp_type
        else:
            media_type.parameters['dcp-type'] = cls._dcp_type_param(media_type, os.path.basename(file_path))
        return media_type

    @staticmethod
    def _dcp_type_param(media_type, filename):
        if media_type.main_type == 'application/json':
            (filename_without_extension, ext) = os.path.splitext(filename)
            if filename_without_extension in DcpMediaType.DCP_METADATA_JSON_FILES:
                return "metadata/{filename}".format(filename=filename_without_extension)
            else:
                return "metadata"
        else:
            return "data"



#!/usr/bin/env python
# coding: utf-8

from __future__ import absolute_import, division, print_function, unicode_literals
import io, os, sys, unittest

pkg_root = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))  # noqa
sys.path.insert(0, pkg_root)  # noqa

from dcplib import s3_multipart
from dcplib.checksumming_io import ChecksummingBufferedReader
from . import TEST_FILE, TEST_FILE_CHECKSUMS


class TestChecksummingBufferedReader(unittest.TestCase):

    file_size = os.path.getsize(TEST_FILE)
    chunk_size = s3_multipart.get_s3_multipart_chunk_size(file_size)

    def check_sums(self, checksums):
        self.assertEqual(checksums['sha1'], TEST_FILE_CHECKSUMS['sha1'])
        self.assertEqual(checksums['sha256'], TEST_FILE_CHECKSUMS['sha256'])
        self.assertEqual(checksums['crc32c'].lower(), TEST_FILE_CHECKSUMS['crc32c'].lower())
        self.assertEqual(checksums['s3_etag'], TEST_FILE_CHECKSUMS['s3_etag'])

    def test_checksums_after_single_read(self):
        with io.open(TEST_FILE, 'rb') as fh:
            reader = ChecksummingBufferedReader(fh, self.chunk_size)
            reader.read()
            sums = reader.get_checksums()
            self.check_sums(sums)

    def test_checksums_after_multiple_reads(self):
        with io.open(TEST_FILE, 'rb') as raw_fh:
            reader = ChecksummingBufferedReader(raw_fh, self.chunk_size)
            while True:
                buf = reader.read(self.chunk_size)
                if not buf:
                    break
            sums = reader.get_checksums()
            self.check_sums(sums)


if __name__ == '__main__':
    unittest.main()



import hashlib
import logging
import os
import struct
import sys

from .s3_etag import S3Etag

logger = logging.getLogger(__name__)
PYTHON_VERSION_CUTOFF = '3.5'

"""
A file-like object that computes checksums for the data written to it, discarding the actual data. Current accepted
checksum functions are crc32c, s3_etag, sha1, and sha256.
"""


class ChecksummingSink:
    class CRC32C:
        def __init__(self, data=None):
            try:
                import crc32c
            except ImportError as e:
                logger.warning("crc32c: %s. Switching to software mode, which may be slow." % e)
                os.environ["CRC32C_SW_MODE"] = "auto"
                import crc32c
            self._checksum_value = crc32c.crc32(data if data is not None else b"")
            self._crc32c = crc32c

        def update(self, data):
            self._checksum_value = self._crc32c.crc32(data, self._checksum_value)

        def hexdigest(self):
            _checksum_byte_value = self._long_to_bytes(self._checksum_value)
            if sys.version < PYTHON_VERSION_CUTOFF:
                import binascii
                return binascii.hexlify(_checksum_byte_value).decode("utf-8")
            else:
                return _checksum_byte_value.hex()

        def _long_to_bytes(self, long_value):
            """
            Turns a long value into its byte string equivalent.
            :param long_value: the long value to be returned as a byte string
            :return: a byte string equivalent of a long value
            """
            _byte_string = b''
            pack = struct.pack
            while long_value > 0:
                _byte_string = pack(b'>I', long_value & 0xffffffff) + _byte_string
                long_value = long_value >> 32
            for i in range(len(_byte_string)):
                if _byte_string[i] != b'\000'[0]:
                    break
            else:
                _byte_string = b'\000'
                i = 0
            _byte_string = _byte_string[i:]

            return _byte_string

    def __init__(self, write_chunk_size, hash_functions=('crc32c', 'sha1', 'sha256', 's3_etag')):
        """ ChecksummingSink is initialized with a map from the name of the hash function to an object that
        represents the calculation function."""
        self._hashers = dict()
        for hasher in hash_functions:
            if hasher == 'crc32c':
                self._hashers['crc32c'] = self.CRC32C()
            elif hasher == 'sha1':
                self._hashers['sha1'] = hashlib.sha1()
            elif hasher == 'sha256':
                self._hashers['sha256'] = hashlib.sha256()
            elif hasher == 's3_etag':
                self._hashers['s3_etag'] = S3Etag(write_chunk_size)

    def write(self, data):
        for hasher in self._hashers.values():
            hasher.update(data)

    def get_checksums(self):
        checksums = {}
        checksums.update({name: hasher.hexdigest() for name, hasher in self._hashers.items()})
        return checksums

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        pass



import boto3
from botocore.errorfactory import ClientError


class AwsSecret:

    """
    Wrapper for AWS secrets.

    Usage:
        secret = AwsSecret(name="my/component/secret")
        secret.update(value='{"foo":"bar"}')
        secret.value
        # -> '{"foo":"bar"}'
        secret.delete()

    Update handles create vs update and undeletion if necessary.
    """

    def __init__(self, name):
        self.name = name
        self.secrets_mgr = boto3.client(service_name='secretsmanager')
        self.secret_metadata = None
        self._load()

    @property
    def value(self):
        if not self.secret_metadata:
            raise RuntimeError("No such secret")
        if 'DeletedDate' in self.secret_metadata:
            raise RuntimeError("This secret is deleted")
        response = self.secrets_mgr.get_secret_value(SecretId=self.secret_metadata['ARN'])
        return response['SecretString']

    def update(self, value):
        if not self.secret_metadata:
            self.secrets_mgr.create_secret(Name=self.name, SecretString=value)
            self._load()
        else:
            self._restore()
            self.secrets_mgr.put_secret_value(SecretId=self.secret_metadata['ARN'],
                                              SecretString=value)

    def delete(self):
        if not self.secret_metadata:
            raise RuntimeError("No such secret")
        if 'DeletedDate' not in self.secret_metadata:
            self.secrets_mgr.delete_secret(SecretId=self.secret_metadata['ARN'])
            self._load()

    def _load(self):
        try:
            response = self.secrets_mgr.describe_secret(SecretId=self.name)
            if response:
                self.secret_metadata = response
        except ClientError as e:
            if e.response['Error']['Code'] != 'ResourceNotFoundException':
                raise e

    def _restore(self):
        if 'DeletedDate' in self.secret_metadata:
            self.secrets_mgr.restore_secret(SecretId=self.secret_metadata['ARN'])
            self._load()



import hashlib

"""
A digest generator that generates S3 ETAGs
with the same interface as the hashlib generators.
"""


class S3Etag:

    def __init__(self, chunk_size):
        self._etag_bytes = 0
        self._etag_parts = []
        self._etag_hasher = hashlib.md5()
        self._chunk_size = chunk_size

    @property
    def chunk_size(self):
        return self._chunk_size

    def update(self, chunk):
        if self._etag_bytes + len(chunk) > self.chunk_size:
            chunk_head = chunk[:self.chunk_size - self._etag_bytes]
            chunk_tail = chunk[self.chunk_size - self._etag_bytes:]
            self._etag_hasher.update(chunk_head)
            self._etag_parts.append(self._etag_hasher.digest())
            self._etag_hasher = hashlib.md5()
            self._etag_hasher.update(chunk_tail)
            self._etag_bytes = len(chunk_tail)
        else:
            self._etag_hasher.update(chunk)
            self._etag_bytes += len(chunk)

    def hexdigest(self):
        if self._etag_bytes:
            self._etag_parts.append(self._etag_hasher.digest())
            self._etag_bytes = 0
        if len(self._etag_parts) > 1:
            etag_csum = hashlib.md5(b"".join(self._etag_parts)).hexdigest()
            return '{}-{}'.format(etag_csum, len(self._etag_parts))
        else:
            return self._etag_hasher.hexdigest()



import hashlib

from io import BufferedReader

import crcmod

from .s3_etag import S3Etag


class ChecksummingBufferedReader:

    def __init__(self, file_handler, read_chunk_size, *args, **kwargs):
        """
        :param file_handler: the file handler to read from
        :param read_file_size: file size for correctly setting the s3 etag chunk size
        """
        self._hashers = dict(crc32c=crcmod.predefined.Crc("crc-32c"),
                             sha1=hashlib.sha1(),
                             sha256=hashlib.sha256(),
                             s3_etag=S3Etag(read_chunk_size))
        self._reader = BufferedReader(file_handler, *args, **kwargs)
        self.raw = self._reader.raw

    def read(self, size=None):
        chunk = self._reader.read(size)
        if chunk:
            for hasher in self._hashers.values():
                hasher.update(chunk)
        return chunk

    def get_checksums(self):
        checksums = {}
        checksums.update({name: hasher.hexdigest() for name, hasher in self._hashers.items()})
        return checksums

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        pass



import os
import unittest

from dcplib.media_types import MediaType


class TestMediaType(unittest.TestCase):

    TEST_STRINGS_AND_ATTRS = {
        'application/json': {
            'top_level_type': 'application', 'subtype': 'json'
        },
        'application/json+zip': {
            'top_level_type': 'application', 'subtype': 'json', 'suffix': 'zip',
        },
        'application/octet-stream+zip; dcp-type=data': {
            'top_level_type': 'application', 'subtype': 'octet-stream',
            'suffix': 'zip', 'parameters': {'dcp-type': 'data'}
        },
        'application/json; dcp-type="metadata/sample"': {
            'top_level_type': 'application', 'subtype': 'json', 'parameters': {'dcp-type': 'metadata/sample'}
        },
        'application/json; dcp-type="metadata/sample"; other-param=foo; quoted-param="bar/baz"': {
            'top_level_type': 'application', 'subtype': 'json',
            'parameters': {'dcp-type': 'metadata/sample', 'other-param': 'foo', 'quoted-param': 'bar/baz'}
        },
        'application/octet-stream; param-with-semicolon-in-value="!#%&~<>@;$*+-"; param2=value2': {
            'top_level_type': 'application', 'subtype': 'octet-stream',
            'parameters': {'param-with-semicolon-in-value': '!#%&~<>@;$*+-', 'param2': 'value2'}
        },
        'application/octet-stream; param-with-escaped-quote-in-value="foo\\"bar"; param2=value2': {
            'top_level_type': 'application', 'subtype': 'octet-stream',
            'parameters': {'param-with-escaped-quote-in-value': 'foo"bar', 'param2': 'value2'}
        }
    }

    def test_string_generation_from_attrs(self):
        for media_type, attributes in self.TEST_STRINGS_AND_ATTRS.items():
            mt = MediaType(**attributes)
            self.assertEqual(media_type, str(mt))

    def test_string_parsing_to_attrs_to_string_generation_round_trip(self):
        for media_type_string, attributes in self.TEST_STRINGS_AND_ATTRS.items():

            mt = MediaType.from_string(media_type_string)

            self.assertEqual(attributes['top_level_type'], mt.top_level_type)
            self.assertEqual(attributes['subtype'], mt.subtype)
            if 'suffix' in attributes:
                self.assertEqual(attributes['suffix'], mt.suffix)
            if 'parameters' in attributes:
                self.assertEqual(attributes['parameters'], mt.parameters)

            self.assertEqual(str(mt), media_type_string)

    BUNDLE_FIXTURE_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), 'fixtures'))

    TEST_FILES_AND_STRINGS = {
        'assay.json': 'application/json',
        'SRR2963332_1.fastq.gz': 'application/gzip',
        'demo.gene.counts.txt': 'text/plain',
        'mystery_file': 'application/octet-stream'
    }

    def test_file_recognition(self):
        for filename, media_type in TestMediaType.TEST_FILES_AND_STRINGS.items():
            file_path = os.path.join(self.BUNDLE_FIXTURE_DIR, filename)
            self.assertEqual(str(MediaType.from_file(file_path)), media_type)



KiB = 1024
MiB = KiB * KiB

AWS_MIN_CHUNK_SIZE = 64 * MiB
"""Files must be larger than this before we consider multipart uploads."""

MULTIPART_THRESHOLD = AWS_MIN_CHUNK_SIZE + 1
"""Convenience variable for Boto3 TransferConfig(multipart_threhold=)."""

AWS_MAX_MULTIPART_COUNT = 10000
"""Maximum number of parts allowed in a multipart upload.  This is a limitation imposed by S3."""


def get_s3_multipart_chunk_size(filesize):
    """Returns the chunk size of the S3 multipart object, given a file's size."""
    if filesize <= AWS_MAX_MULTIPART_COUNT * AWS_MIN_CHUNK_SIZE:
        return AWS_MIN_CHUNK_SIZE
    else:
        div = filesize // AWS_MAX_MULTIPART_COUNT
        if div * AWS_MAX_MULTIPART_COUNT < filesize:
            div += 1
        return ((div + MiB - 1) // MiB) * MiB



import logging

try:
    from threading import get_ident
except ImportError:
    from thread import get_ident

import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util import retry, timeout

logger = logging.getLogger(__name__)

class RetryPolicy(retry.Retry):
    def __init__(self, retry_after_status_codes={301}, *args, **kwargs):
        super(RetryPolicy, self).__init__(*args, **kwargs)
        self.RETRY_AFTER_STATUS_CODES = frozenset(retry_after_status_codes | retry.Retry.RETRY_AFTER_STATUS_CODES)

    def increment(self, *args, **kwargs):
        retry = super(RetryPolicy, self).increment(*args, **kwargs)
        logger.warning("Retrying: {}".format(retry.history[-1]))
        return retry

class HTTPRequest:
    """
    A requests wrapper preconfigured with best practice timeout and retry policies and per-thread session tracking.

    The timeout and retry policies are configured to turn on exponential backoff support provided by urllib3, and to
    make the client resistant to both network issues and intermittent server-side errors.

    The per-thread session tracking is to avoid many threads sharing the same session in multithreaded environments.
    """
    retry_policy = RetryPolicy(read=4,
                               status=4,
                               backoff_factor=0.1,
                               status_forcelist=frozenset({500, 502, 503, 504}))
    timeout_policy = timeout.Timeout(connect=20, read=40)

    def __init__(self):
        self.sessions = {}

    def __call__(self, *args, **kwargs):
        if get_ident() not in self.sessions:
            session = requests.Session()
            adapter = HTTPAdapter(max_retries=self.retry_policy)
            session.mount('http://', adapter)
            session.mount('https://', adapter)
            self.sessions[get_ident()] = session
        return self.sessions[get_ident()].request(*args, timeout=self.timeout_policy, **kwargs)

    def head(self, url, *args, **kwargs):
        return self(url=url, method="HEAD", *args, **kwargs)

    def get(self, url, *args, **kwargs):
        return self(url=url, method="GET", *args, **kwargs)

    def put(self, url, *args, **kwargs):
        return self(url=url, method="PUT", *args, **kwargs)

    def post(self, url, *args, **kwargs):
        return self(url=url, method="POST", *args, **kwargs)

    def delete(self, url, *args, **kwargs):
        return self(url=url, method="DELETE", *args, **kwargs)

    def patch(self, url, *args, **kwargs):
        return self(url=url, method="PATCH", *args, **kwargs)

    def options(self, url, *args, **kwargs):
        return self(url=url, method="OPTIONS", *args, **kwargs)

http = HTTPRequest()



import json
import botocore
from . import clients, resources


class ARN:
    fields = "arn partition service region account_id resource".split()
    _default_region, _default_account_id, _default_iam_username = None, None, None

    def __init__(self, arn="arn:aws::::", **kwargs):
        self.__dict__.update(dict(zip(self.fields, arn.split(":", 5)), **kwargs))
        if "region" not in kwargs and not self.region:
            self.region = self.get_region()
        if "account_id" not in kwargs and not self.account_id:
            self.account_id = self.get_account_id()

    @classmethod
    def get_region(cls):
        if cls._default_region is None:
            cls._default_region = botocore.session.Session().get_config_variable("region")
        return cls._default_region

    @classmethod
    def get_account_id(cls):
        if cls._default_account_id is None:
            cls._default_account_id = clients.sts.get_caller_identity()["Account"]
        return cls._default_account_id

    def __str__(self):
        return ":".join(getattr(self, field) for field in self.fields)


def send_sns_msg(topic_arn, message, attributes=None):
    sns_topic = resources.sns.Topic(str(topic_arn))
    args = {'Message': json.dumps(message)}
    if attributes is not None:
        args['MessageAttributes'] = attributes
    sns_topic.publish(**args)
